/******************************************************************************/
/* SlunkCrypt, by LoRd_MuldeR <MuldeR2@GMX.de>                                */
/* This work has been released under the CC0 1.0 Universal license!           */
/******************************************************************************/

#ifdef _MSC_VER
#define PTW32_STATIC_LIB 1
#endif

/* Internal */
#include "slunkcrypt.h"
#include "thread.h"

/* CRT */
#include <stdlib.h>

/* PThread */
#include <pthread.h>
#include <semaphore.h>

typedef struct
{
	thrdpl_worker_t worker;
	void *args;
}
thrdpl_task_t;

typedef struct
{
	size_t index;
	pthread_mutex_t *mutex;
	pthread_cond_t *ready;
	size_t *queue;
	sem_t sem_free, sem_used;
	pthread_t thread;
	thrdpl_task_t task;
}
thrdpl_thread_t;

typedef struct
{
	size_t thread_count, index, queue;
	pthread_mutex_t mutex;
	pthread_cond_t ready;
	thrdpl_thread_t threads[MAX_THREADS];
}
thrdpl_data_t;

// ==========================================================================
// Thread main
// ==========================================================================

static void *thread_main(void *const arg)
{
	thrdpl_thread_t *const data = (thrdpl_thread_t*)arg;

	for (;;)
	{
		if (sem_wait(&data->sem_used) != 0)
		{
			abort();
		}

		if (pthread_mutex_lock(data->mutex) != 0)
		{
			abort();
		}

		const thrdpl_worker_t worker = data->task.worker;
		void *const args = data->task.args;

		if (pthread_mutex_unlock(data->mutex) != 0)
		{
			abort();
		}

		worker(args);

		if (pthread_mutex_lock(data->mutex) != 0)
		{
			abort();
		}

		if (!(*data->queue -= 1U))
		{
			if (pthread_cond_broadcast(data->ready) != 0)
			{
				abort();
			}
		}

		if (pthread_mutex_unlock(data->mutex) != 0)
		{
			abort();
		}

		if (sem_post(&data->sem_free) != 0)
		{
			abort();
		}
	}
}

// ==========================================================================
// Manage threads
// ==========================================================================

static int create_thread(const size_t index, thrdpl_thread_t *const thread_data, pthread_mutex_t *const mutex, pthread_cond_t *const ready, size_t *const queue)
{
	thread_data->index = index;
	thread_data->mutex = mutex;
	thread_data->ready = ready;
	thread_data->queue = queue;

	if (sem_init(&thread_data->sem_free, 0, 1U) != 0)
	{
		return -1;
	}

	if (sem_init(&thread_data->sem_used, 0, 0U) != 0)
	{
		sem_destroy(&thread_data->sem_free);
		return -1;
	}

	if (pthread_create(&thread_data->thread, NULL, thread_main, thread_data) != 0)
	{
		sem_destroy(&thread_data->sem_used);
		sem_destroy(&thread_data->sem_free);
		return -1;
	}

	return 0;
}

static int destroy_thread(thrdpl_thread_t *const thread_data)
{
	pthread_cancel(thread_data->thread);
	pthread_join(thread_data->thread, NULL);
	sem_destroy(&thread_data->sem_used);
	sem_destroy(&thread_data->sem_free);
	return 0;
}

// ==========================================================================
// Thread pool API
// ==========================================================================

thrdpl_t thrdpl_create(const size_t count)
{
	size_t i, j;
	thrdpl_data_t *pool = NULL;

	if ((count < 1U) || (count > MAX_THREADS))
	{
		return THRDPL_NULL;
	}
	
	if (!(pool = (thrdpl_data_t*) malloc(sizeof(thrdpl_data_t))))
	{
		return THRDPL_NULL;
	}

	slunkcrypt_bzero(pool, sizeof(thrdpl_data_t));
	pool->thread_count = count;

	if (pthread_mutex_init(&pool->mutex, NULL) != 0)
	{
		goto failure;
	}

	if (pthread_cond_init(&pool->ready, NULL) != 0)
	{
		pthread_mutex_destroy(&pool->mutex);
		goto failure;
	}

	for (i = 0U; i < count; ++i)
	{
		if (create_thread(i, &pool->threads[i], &pool->mutex, &pool->ready, &pool->queue) != 0)
		{
			for (j = 0U; j < i; ++j)
			{
				destroy_thread(&pool->threads[j]);
			}
			pthread_cond_destroy(&pool->ready);
			pthread_mutex_destroy(&pool->mutex);
			goto failure;
		}
	}

	return (thrdpl_t)pool;

failure:
	free(pool);
	return (thrdpl_t)NULL;
}

void thrdpl_submit(const thrdpl_t thrdpl, const thrdpl_worker_t worker, void *const args)
{
	thrdpl_data_t *const pool = (thrdpl_data_t*)thrdpl;
	
	if (pthread_mutex_lock(&pool->mutex) != 0)
	{
		abort();
	}

	thrdpl_thread_t *const thread = &pool->threads[pool->index];

	if (++pool->index >= pool->thread_count)
	{
		pool->index = 0U;
	}

	++pool->queue;

	if (pthread_mutex_unlock(&pool->mutex) != 0)
	{
		abort();
	}

	if (sem_wait(&thread->sem_free) != 0)
	{
		abort();
	}

	thread->task.worker = worker;
	thread->task.args = args;
	
	if (sem_post(&thread->sem_used) != 0)
	{
		abort();
	}
}

void thrdpl_await(const thrdpl_t thrdpl)
{
	thrdpl_data_t *const pool = (thrdpl_data_t*)thrdpl;

	if (pthread_mutex_lock(&pool->mutex) != 0)
	{
		abort();
	}

	while (pool->queue)
	{
		if (pthread_cond_wait(&pool->ready, &pool->mutex) != 0)
		{
			abort();
		}
	}

	if (pthread_mutex_unlock(&pool->mutex) != 0)
	{
		abort();
	}
}

void thrdpl_destroy(const thrdpl_t thrdpl)
{
	size_t i;
	thrdpl_data_t *const pool = (thrdpl_data_t*)thrdpl;
	if (pool)
	{
		for (i = 0U; i < pool->thread_count; ++i)
		{
			destroy_thread(&pool->threads[i]);
		}
		pthread_cond_destroy(&pool->ready);
		pthread_mutex_destroy(&pool->mutex);
		free(pool);
	}
}
